import re

from six import string_types

from ksql.errors import (
    SQLTypeNotImplementYetError,
    IllegalTableTypeError,
    IllegalValueFormatError,
    SQLFormatNotImplementError,
    BuildNotImplmentError,
)


class SQLBuilder(object):
    @classmethod
    def build(self, sql_type, **kwargs):
        if sql_type == "create":
            sql_builder = CreateBuilder(kwargs.pop("table_type"))
            sql_str = sql_builder.build(
                kwargs.pop("table_name"),
                kwargs.pop("columns_type"),
                kwargs.pop("topic"),
                kwargs.pop("value_format", "JSON"),
                kwargs.pop("key", None),
            )

        elif sql_type == "create_as":
            sql_builder = CreateAsBuilder(kwargs.pop("table_type"))
            sql_str = sql_builder.build(
                table_name=kwargs.pop("table_name"),
                select_columns=kwargs.pop("select_columns", None),
                src_table=kwargs.pop("src_table"),
                kafka_topic=kwargs.pop("kafka_topic"),
                value_format=kwargs.pop("value_format"),
                partition_by=kwargs.pop("partition_by", None),
                **kwargs,
            )

        else:
            raise SQLTypeNotImplementYetError(sql_type)

        return sql_str


class BaseCreateBuilder(object):
    def __init__(self, table_type, sql_format=None):
        self.table_types = ["table", "stream"]
        self.value_formats = ["delimited", "json", "avro"]
        self.table_type = table_type
        self.sql_format = sql_format

        if table_type.lower() not in self.table_types:
            raise IllegalTableTypeError(table_type)

        if not sql_format:
            raise SQLFormatNotImplementError()

    def build(self):
        raise BuildNotImplmentError()

    def _parsed_with_properties(self, **kwargs):
        properties = [""]
        for prop in self.properties:
            if prop in kwargs.keys():
                value = kwargs.get(prop)
                if isinstance(value, string_types):
                    p = "{}='{}'".format(prop, kwargs.get(prop))
                else:
                    p = "{}={}".format(prop, kwargs.get(prop))

                properties.append(p)

        return ", ".join(properties)


class CreateBuilder(BaseCreateBuilder):
    def __init__(self, table_type):
        str_format = "CREATE {} {} ({}) WITH (kafka_topic='{}', value_format='{}'{});"
        super(CreateBuilder, self).__init__(table_type, str_format)
        self.properties = ["kafka_topic", "value_format", "key", "timestamp"]

    def build(self, table_name, columns_type=[], topic=None, value_format="JSON", key=None):
        if value_format.lower() not in self.value_formats:
            raise IllegalValueFormatError(value_format)

        built_colums_type = self._build_colums_type(columns_type)
        built_key = self._build_key(key)

        sql_str = self.sql_format.format(self.table_type, table_name, built_colums_type, topic, value_format, built_key)
        return sql_str

    @staticmethod
    def _build_colums_type(columns_type):
        built_columns_type = ", ".join(columns_type)
        return built_columns_type

    @staticmethod
    def _build_key(key):
        if key:
            built_key = ", key='{}'".format(key)
            return built_key
        else:
            return ""


class CreateAsBuilder(BaseCreateBuilder):
    def __init__(self, table_type):
        str_format = "CREATE {} {} WITH (kafka_topic='{}', value_format='{}'{}) AS SELECT {} FROM {} {} {}"
        super(CreateAsBuilder, self).__init__(table_type, str_format)
        self.properties = ["kafka_topic", "value_format", "partitions", "replicas", "timestamp"]

    def build(
        self,
        table_name,
        select_columns,
        src_table,
        kafka_topic=None,
        value_format="JSON",
        conditions=[],
        partition_by=None,
        **kwargs
    ):

        if value_format.lower() not in self.value_formats:
            raise IllegalValueFormatError(value_format)

        if not kafka_topic:
            kafka_topic = table_name

        select_clause, where_clause, partition_by_clause, properties = self._build_clauses(
            select_columns, conditions, partition_by, **kwargs
        )

        sql_str = self.sql_format.format(
            self.table_type,
            table_name,
            kafka_topic,
            value_format,
            properties,
            select_clause,
            src_table,
            where_clause,
            partition_by_clause,
        )

        cleaned_sql_str = re.sub(r"\s+", " ", sql_str).strip()

        return cleaned_sql_str

    def _build_clauses(self, select_columns, conditions, partition_by, **kwargs):
        select_clause = self._build_select_clause(select_columns)
        where_clause = self._build_where_clause(conditions)
        partition_by_clause = self._build_partition_by_clause(partition_by)
        properties = self._parsed_with_properties(**kwargs)

        return select_clause, where_clause, partition_by_clause, properties

    @staticmethod
    def _build_where_clause(conditions):
        if len(conditions) > 0:
            where_clause = "where {}".format(conditions).replace('"', "'")
            return where_clause
        else:
            return ""

    @staticmethod
    def _build_select_clause(select_columns):
        select_clause = "*"
        if select_columns:
            if len(select_columns) > 0:
                select_clause = ", ".join(select_columns)

        return select_clause

    @staticmethod
    def _build_partition_by_clause(partition_by):
        if partition_by:
            return "PARTITION BY {}".format(partition_by)
        else:
            return ""



import os

from ksql.client import KSQLAPI
from ksql.errors import FileTypeError, InvalidQueryError


class FileUpload(object):
    """ UPLOAD KSQL RULES AS FILES """

    def __init__(self, url, **kwargs):
        """ Instantiate the url pointer and the client object """
        self.url = url
        self.client = KSQLAPI(url)

    def upload(self, ksqlfile):
        """
        A method to upload ksql rules as .ksql file

        Parameter List
        -------------
        :param ksqlfile: File containing the ksql rules to be uploaded.
                          Only supports ksql queries and not streaming queries

        """

        # check if the file is .ksql
        self.checkExtension(ksqlfile)

        # parse the file and get back the rules
        rules = self.get_rules_list(ksqlfile)
        log_return = []
        for each_rule in rules:
            resp = self.client.ksql(each_rule)
            log_return.append(resp)

        return log_return

    def get_rules_list(self, ksqlfile):
        with open(ksqlfile) as rf:
            rule = ""

            for line in rf:
                rule = rule + " " + line.strip()

                if rule[-1:] == ";":
                    yield rule
                    rule = ""

            if rule[-1:] != ";":
                raise InvalidQueryError(rule)

    def checkExtension(self, filename):
        ext = os.path.splitext(filename)[-1].lower()

        if ext != ".ksql":
            raise FileTypeError(ext)
        return



import time

import base64
import functools
import json
import logging
import requests
import urllib
from copy import deepcopy
from requests import Timeout

from ksql.builder import SQLBuilder
from ksql.errors import CreateError, InvalidQueryError, KSQLError


class BaseAPI(object):
    def __init__(self, url, **kwargs):
        self.url = url
        self.max_retries = kwargs.get("max_retries", 3)
        self.delay = kwargs.get("delay", 0)
        self.timeout = kwargs.get("timeout", 15)
        self.api_key = kwargs.get("api_key")
        self.secret = kwargs.get("secret")
        self.headers = {
            'Content-Type': 'application/vnd.ksql.v1+json; charset=utf-8',
        }

    def get_timout(self):
        return self.timeout

    @staticmethod
    def _validate_sql_string(sql_string):
        if len(sql_string) > 0:
            if sql_string[-1] != ";":
                sql_string = sql_string + ";"
        else:
            raise InvalidQueryError(sql_string)
        return sql_string

    @staticmethod
    def _raise_for_status(r, response):
        r_json = json.loads(response)
        if r.getcode() != 200:
            # seems to be the new API behavior
            if r_json.get("@type") == "statement_error" or r_json.get("@type") == "generic_error":
                error_message = r_json["message"]
                error_code = r_json["error_code"]
                stackTrace = r_json["stack_trace"]
                raise KSQLError(error_message, error_code, stackTrace)
            else:
                raise KSQLError("Unknown Error: {}".format(r.content))
        else:
            # seems to be the old API behavior, so some errors have status 200, bug??
            if r_json[0]["@type"] == "currentStatus" and r_json[0]["commandStatus"]["status"] == "ERROR":
                error_message = r_json[0]["commandStatus"]["message"]
                error_code = None
                stackTrace = None
                raise KSQLError(error_message, error_code, stackTrace)
            return True

    def ksql(self, ksql_string, stream_properties=None):
        r = self._request(endpoint="ksql", sql_string=ksql_string, stream_properties=stream_properties)
        response = r.read().decode("utf-8")
        self._raise_for_status(r, response)
        res = json.loads(response)
        return res

    def query(self, query_string, encoding="utf-8", chunk_size=128, stream_properties=None, idle_timeout=None):
        """
        Process streaming incoming data.

        """
        streaming_response = self._request(
            endpoint="query", sql_string=query_string, stream_properties=stream_properties
        )
        start_idle = None

        if streaming_response.code == 200:
            for chunk in streaming_response:
                if chunk != b"\n":
                    start_idle = None
                    yield chunk.decode(encoding)
                else:
                    if not start_idle:
                        start_idle = time.time()
                    if idle_timeout and time.time() - start_idle > idle_timeout:
                        print("Ending query because of time out! ({} seconds)".format(idle_timeout))
                        return
        else:
            raise ValueError("Return code is {}.".format(streaming_response.status_code))

    def get_request(self, endpoint):
        auth = (self.api_key, self.secret) if self.api_key or self.secret else None
        return requests.get(endpoint, headers=self.headers, auth=auth)

    def _request(self, endpoint, method="POST", sql_string="", stream_properties=None, encoding="utf-8"):
        url = "{}/{}".format(self.url, endpoint)

        logging.debug("KSQL generated: {}".format(sql_string))

        sql_string = self._validate_sql_string(sql_string)
        body = {"ksql": sql_string}
        if stream_properties:
            body["streamsProperties"] = stream_properties
        else:
            body["streamsProperties"] = {}
        data = json.dumps(body).encode(encoding)

        headers = deepcopy(self.headers)
        if self.api_key and self.secret:
            base64string = base64.b64encode(bytes("{}:{}".format(self.api_key, self.secret), "utf-8")).decode("utf-8")
            headers["Authorization"] = "Basic %s" % base64string

        req = urllib.request.Request(url=url, data=data, headers=headers, method=method.upper())

        try:
            r = urllib.request.urlopen(req, timeout=self.timeout)
        except urllib.error.HTTPError as http_error:
            try:
                content = json.loads(http_error.read().decode(encoding))
            except Exception as e:
                raise http_error
            else:
                logging.debug("content: {}".format(content))
                raise KSQLError(e=content.get("message"), error_code=content.get("error_code"))
        else:
            return r

    @staticmethod
    def retry(exceptions, delay=1, max_retries=5):
        """
        A decorator for retrying a function call with a specified delay in case of a set of exceptions

        Parameter List
        -------------
        :param exceptions:  A tuple of all exceptions that need to be caught for retry
                                            e.g. retry(exception_list = (Timeout, Readtimeout))
        :param delay: Amount of delay (seconds) needed between successive retries.
        :param times: no of times the function should be retried

        """

        def outer_wrapper(function):
            @functools.wraps(function)
            def inner_wrapper(*args, **kwargs):
                final_excep = None
                for counter in range(max_retries):
                    if counter > 0:
                        time.sleep(delay)
                    final_excep = None
                    try:
                        value = function(*args, **kwargs)
                        return value
                    except (exceptions) as e:
                        final_excep = e
                        pass  # or log it

                if final_excep is not None:
                    raise final_excep

            return inner_wrapper

        return outer_wrapper


class SimplifiedAPI(BaseAPI):
    def __init__(self, url, **kwargs):
        super(SimplifiedAPI, self).__init__(url, **kwargs)

    def create_stream(self, table_name, columns_type, topic, value_format="JSON"):
        return self._create(
            table_type="stream",
            table_name=table_name,
            columns_type=columns_type,
            topic=topic,
            value_format=value_format,
        )

    def create_table(self, table_name, columns_type, topic, value_format, key):
        if not key:
            raise ValueError("key is required for creating a table.")
        return self._create(
            table_type="table",
            table_name=table_name,
            columns_type=columns_type,
            topic=topic,
            value_format=value_format,
            key=key,
        )

    def create_stream_as(
        self,
        table_name,
        select_columns,
        src_table,
        kafka_topic=None,
        value_format="JSON",
        conditions=[],
        partition_by=None,
        **kwargs
    ):
        return self._create_as(
            table_type="stream",
            table_name=table_name,
            select_columns=select_columns,
            src_table=src_table,
            kafka_topic=kafka_topic,
            value_format=value_format,
            conditions=conditions,
            partition_by=partition_by,
            **kwargs,
        )

    def _create(self, table_type, table_name, columns_type, topic, value_format="JSON", key=None):
        ksql_string = SQLBuilder.build(
            sql_type="create",
            table_type=table_type,
            table_name=table_name,
            columns_type=columns_type,
            topic=topic,
            value_format=value_format,
            key=key,
        )
        self.ksql(ksql_string)
        return True

    @BaseAPI.retry(exceptions=(Timeout, CreateError))
    def _create_as(
        self,
        table_type,
        table_name,
        select_columns,
        src_table,
        kafka_topic=None,
        value_format="JSON",
        conditions=[],
        partition_by=None,
        **kwargs
    ):
        ksql_string = SQLBuilder.build(
            sql_type="create_as",
            table_type=table_type,
            table_name=table_name,
            select_columns=select_columns,
            src_table=src_table,
            kafka_topic=kafka_topic,
            value_format=value_format,
            conditions=conditions,
            partition_by=partition_by,
            **kwargs,
        )
        self.ksql(ksql_string)
        return True



import ksql
import telnetlib


def check_kafka_available(bootstrap_servers):
    host, port = bootstrap_servers.split(":")
    try:
        telnetlib.Telnet(host, port)
        return True
    except Exception:
        return False


def get_all_streams(api_client, prefix=None):
    all_streams = api_client.ksql("""SHOW STREAMS;""")
    filtered_streams = []
    for stream in all_streams[0]["streams"]:
        if stream["type"] != "STREAM":
            continue
        if prefix and not stream["name"].startswith(prefix.upper()):
            continue
        filtered_streams.append(stream["name"])
    return filtered_streams


def get_stream_info(api_client, stream_name):
    try:
        r = api_client.ksql("""DESCRIBE EXTENDED {}""".format(stream_name))
    except ksql.errors.KSQLError as e:
        if e.error_code == 40001:
            return None
        else:
            raise
    stream_info = r[0]["sourceDescription"]
    return stream_info


def drop_all_streams(api_client, prefix=None):
    filtered_streams = get_all_streams(api_client, prefix=prefix)
    for stream in filtered_streams:
        drop_stream(api_client, stream)


def drop_stream(api_client, stream_name):
    read_queries, write_queries = get_dependent_queries(api_client, stream_name)
    dependent_queries = read_queries + write_queries
    for query in dependent_queries:
        api_client.ksql("""TERMINATE {};""".format(query))
    api_client.ksql(
        """DROP
    STREAM IF EXISTS
    {};""".format(
            stream_name
        )
    )


def get_dependent_queries(api_client, stream_name):
    stream_info = get_stream_info(api_client, stream_name)
    read_queries = []
    write_queries = []
    if stream_info and stream_info["readQueries"]:
        read_queries = [query["id"] for query in stream_info["readQueries"]]
    if stream_info and stream_info["writeQueries"]:
        write_queries = [query["id"] for query in stream_info["writeQueries"]]

    return read_queries, write_queries



class SQLTypeNotImplementYetError(Exception):
    def __init__(self, sql_type):
        self.msg = "This type {} has not be implement yet.".format(sql_type)


class IllegalTableTypeError(Exception):
    def __init__(self, table_type):
        self.msg = "This table type {} is illegal.".format(table_type)


class IllegalValueFormatError(Exception):
    def __init__(self, value_format):
        self.msg = "This value format {} is illegal.".format(value_format)


class CreateError(Exception):
    def __init__(self, e):
        self.msg = "{}".format(e)


class SQLFormatNotImplementError(Exception):
    pass


class BuildNotImplmentError(Exception):
    pass


class FileTypeError(Exception):
    def __init__(self, ext):
        self.msg = "This {} file extension is not valid".format(ext)


class InvalidQueryError(Exception):
    def __init__(self, query):
        self.msg = "The query:\n{}\n is invalid".format(query)


class KSQLError(Exception):
    def __init__(self, e, error_code=None, stackTrace=None):
        self.msg = "{}".format(e)
        self.error_code = error_code
        self.stackTrace = stackTrace



from __future__ import absolute_import
from __future__ import print_function

from ksql.api import SimplifiedAPI


class KSQLAPI(object):
    """ API Class """

    def __init__(self, url, max_retries=3, check_version=True, **kwargs):
        """
        You can use a Basic Authentication with this API, for now we accept the api_key/secret based on the Confluent
        Cloud implementation. So you just need to put on the kwargs the api_key and secret.
        """
        self.url = url
        self.sa = SimplifiedAPI(url, max_retries=max_retries, **kwargs)
        if check_version is True:
            self.get_ksql_version()

    def get_url(self):
        return self.url

    @property
    def timeout(self):
        return self.sa.get_timout()

    def get_ksql_version(self):
        r = self.sa.get_request(self.url + "/info")
        if r.status_code == 200:
            info = r.json().get("KsqlServerInfo")
            version = info.get("version")
            return version

        else:
            raise ValueError("Status Code: {}.\nMessage: {}".format(r.status_code, r.content))

    def get_properties(self):
        properties = self.sa.ksql("show properties;")
        return properties[0]["properties"]

    def ksql(self, ksql_string, stream_properties=None):
        return self.sa.ksql(ksql_string, stream_properties=stream_properties)

    def query(self, query_string, encoding="utf-8", chunk_size=128, stream_properties=None, idle_timeout=None):
        return self.sa.query(
            query_string=query_string,
            encoding=encoding,
            chunk_size=chunk_size,
            stream_properties=stream_properties,
            idle_timeout=idle_timeout,
        )

    def create_stream(self, table_name, columns_type, topic, value_format="JSON"):
        return self.sa.create_stream(
            table_name=table_name, columns_type=columns_type, topic=topic, value_format=value_format
        )

    def create_table(self, table_name, columns_type, topic, value_format, key, **kwargs):
        return self.sa.create_table(
            table_name=table_name, columns_type=columns_type, topic=topic, value_format=value_format, key=key, **kwargs
        )

    def create_stream_as(
        self,
        table_name,
        select_columns,
        src_table,
        kafka_topic=None,
        value_format="JSON",
        conditions=[],
        partition_by=None,
        **kwargs
    ):

        return self.sa.create_stream_as(
            table_name=table_name,
            select_columns=select_columns,
            src_table=src_table,
            kafka_topic=kafka_topic,
            value_format=value_format,
            conditions=conditions,
            partition_by=partition_by,
            **kwargs,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Setup module """
import os

from setuptools import setup
from distutils.version import LooseVersion
import pip

if LooseVersion(pip.__version__) >= "10.0.0":
    from pip._internal.req import parse_requirements
else:
    from pip.req import parse_requirements

def get_install_requirements(path):
    content = open(os.path.join(os.path.dirname(__file__), path)).read()
    return [
        req
        for req in content.split("\n")
        if req != '' and not req.startswith('#')
    ]

# Get version from __init__.py file
VERSION = "0.10.1.1"

here = os.path.dirname(__file__)

# Get long description
README = open(os.path.join(os.path.dirname(__file__), "README.rst")).read()

setuptools_kwargs = {
    'install_requires': [
	    'requests',
		'six',
		'urllib3'
    ],
    'zip_safe': False,
}

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="ksql",
    version=VERSION,
    description="A Python wrapper for the KSQL REST API",
    long_description=README,
    author="Bryan Yang",
    author_email="kenshin200528@gmail.com",
    url="https://github.com/bryanyang0528/ksql-python",
    license="MIT License",
    packages=[
        "ksql"
    ],
    include_package_data=True,
    platforms=['any'],
    extras_require={
        "dev": get_install_requirements("test-requirements.txt")
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    **setuptools_kwargs
)
